import os
import shutil

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class HyphenateDistBuildHook(BuildHookInterface):
    PLUGIN_NAME = 'hyphenated-dist-directory'

    def finalize(self, version, build_data, artifact_path):
        self._copy_to_dist_hyphenated(artifact_path)

    def _copy_to_dist_hyphenated(self, artifact_path):
        """Copy the contents in the 'dist' directory, renaming filenames with hyphens instead of underscores.

        > This enables using hyphenated package naming when using pypi.uoregon.edu.

        Args:
            artifact_path (str): The path to the 'dist' directory that will be copied.
        """
        dist_dir = f'{os.path.dirname(artifact_path)}-hyphenated'
        os.makedirs(dist_dir, exist_ok=True)
        dist_filename = os.path.basename(artifact_path)
        dist_filename = dist_filename.replace('_','-')
        new_artifact_path = os.path.join(dist_dir, dist_filename)
        shutil.copyfile(artifact_path, new_artifact_path)
        print(f'Final hyphenated package path: {new_artifact_path}')




from warnings import warn

from uologging.cli import add_verbosity_flag, get_default_parser
from uologging.downloads import DownloadTracer
from uologging.performance import trace
from uologging.uologging import init_console, init_syslog, set_verbosity


def function_renamed(function, old_name):
    """Want to change the name of some function in your python library?
    But you don't want to break your users' code?

    This function will wrap any function with a deprecated alias.
    
    Example:

    Imagine you have the 'say_hello' function in your library.

        >>> def say_hello():
        ...     print('Hello, world!')
    
    You want to change the name to 'hello,' and so you do!

        >>> def hello():
        ...     print('Hello, world!')

    But, you don't want to break your users' code.

    
        >>> say_hello = function_renamed(hello, 'say_hello')
        >>> __all__ = ['hello', 'say_hello']


    Args:
        new_function (_type_): _description_
        deprecated_function_name (_type_): _description_
    """
    def deprecated_function(*args, **kwargs):
        deprecation_warning = f'{old_name}() is deprecated (we plan to remove it in next major-version). Please use {function.__name__}() instead.'
        warn(deprecation_warning, DeprecationWarning, stacklevel=2)
        # NOW, only after warning, pass on to the underlying function
        function(*args, **kwargs)
    return deprecated_function


init_console_logging = function_renamed(init_console, old_name='init_console_logging')
init_syslog_logging = function_renamed(init_syslog, old_name='init_syslog_logging')
set_logging_verbosity = function_renamed(set_verbosity, old_name='set_logging_verbosity')

__version__ = '0.8.0'
__all__ = [
    'add_verbosity_flag', 
    'DownloadTracer',
    'get_default_parser', 
    'init_console_logging', 
    'init_console', 
    'init_syslog_logging',
    'init_syslog',
    'set_logging_verbosity', 
    'set_verbosity',
    'trace',
]



# hello.py
import logging

import uologging

logger = logging.getLogger(__name__)


@uologging.trace(logger)
def hello():
    print('hello from example!')
    logger.debug('This is a DEBUG msg.')
    logger.info('This is a INFO msg.')
    logger.warning('This is a WARNING msg.')



import logging
import sys
import time
from functools import wraps

MAJOR = 0
MINOR = 1


def pretty_function_with_args_str(func, args, kwargs):
    return f'{func.__module__}:{func.__name__}({pretty_args(args, kwargs)})'


def pretty_args(args, kwargs):
    strings = []
    if args:
        strings.extend([repr(arg) for arg in args])
    if kwargs:
        strings.extend([f'{key}={repr(val)}' for key, val in kwargs.items()])
    return ', '.join(strings)


def pretty_function_no_args_str(func):
    return f'{func.__module__}:{func.__name__}(...)'


def trace(logger: logging.Logger, capture_args=True):
    """Decorator to trace the execution time of a Python function or method.

    Uses INFO log level for logging the traced function's "exec time".
    Also will log the traced function's entry and exit at DEBUG log level.

    Args:
        logger (logging.Logger): The logger instance for the function's module.
        capture_args (bool, optional): When tracing the function, capture the 
            arguments provided to the function and add them to the log message.
            Defaults to True.

    Example:

    First, get the logger for your Python module.

        import logging
        logger = logging.getLogger(__name__)

    Then, use the trace_time decorator with that logger as the argument.

        import uologging
        @uologging.trace(logger)
        def my_slow_function():
            import time
            time.sleep(1)
        my_slow_function()
    """
    def _trace_time(func):
        def log_info(msg):
            if sys.version_info[MAJOR] >= 3 and sys.version_info[MINOR] >= 8:
                logger.info(msg, stacklevel=3)
            else:
                logger.info(msg)
        def log_debug(msg):
            if sys.version_info[MAJOR] >= 3 and sys.version_info[MINOR] >= 8:
                logger.info(msg, stacklevel=3)
            else:
                logger.info(msg)
        @wraps(func)
        def timed(*args, **kwargs):
            if capture_args:
                function_str = pretty_function_with_args_str(func, args, kwargs)
            else:
                function_str = pretty_function_no_args_str(func)
            log_debug(f'Starting: {function_str}')
            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()
            log_debug(f'Finished: {function_str}')
            log_info(f'{function_str} exec time: {end - start:.2f} sec')
            return result
        return timed
    return _trace_time



import logging
import logging.handlers

_logging_format = logging.Formatter(
    '%(asctime)s %(levelname)-8s  %(message)s [%(pathname)s:%(lineno)d]',
    '%Y-%m-%d %H:%M:%S')


initialized_syslog_packages = []
def init_syslog(name:str = None):
    """Setup logging to be output to syslog (via /dev/log).

    Args:
        name (str, optional): Which logger to configure. Default to root logger.

    Example:
        Call this method.
        >>> import uologging
        >>> uologging.init_syslog()

        Then use the Python logging package in each of your package's modules.
        >>> import logging
        >>> logger = logging.getLogger(__name__)

        We use a hardcoded str here to enable this doctest:
        >>> logger = logging.getLogger('examplepkg.just.testing')
        >>> logger.critical('Just kidding, this is a test!')    
    """
    global initialized_syslog_packages
    if name not in initialized_syslog_packages:
        syslog_handler = logging.handlers.SysLogHandler(address='/dev/log')
        _add_log_handler_with_formatter(syslog_handler, name)
        initialized_syslog_packages.append(name)


initialized_console_packages = []
def init_console(name:str = None):
    """Setup logging to be output to the console.

    Args:
        name (str, optional): Which logger to configure. Default to root logger.

    Example:
        Call this method.
        >>> import uologging
        >>> uologging.init_console()

        Then use the Python logging package in each of your package's modules.
        >>> import logging
        >>> logger = logging.getLogger(__name__)

        We use a hardcoded str here to enable doctest:
        >>> logger = logging.getLogger('examplepkg.just.testing')
        >>> logger.critical('Just kidding, this is a test!')
    """
    global initialized_console_packages
    if name not in initialized_console_packages:
        console_handler = logging.StreamHandler()
        _add_log_handler_with_formatter(console_handler, name)
        initialized_console_packages.append(name)


def set_verbosity(verbosity_flag: int, name:str = None):
    """Set the logging verbosity for a logger.

    Args:
        verbosity_flag (int): Higher number means more logging. Choices are [0,2]. 
            Default is 0. Default will captures WARNING, ERROR, and CRITICAL logs.
            Provide 1 to also capture INFO logs. Provide 2 to also capture DEBUG logs.
        name (str, optional): Which logger to configure. Default to root logger.
    """
    logger = logging.getLogger(name)
    if verbosity_flag == 1:
        logger.setLevel(logging.INFO)
    elif verbosity_flag >= 2:
        logger.setLevel(logging.DEBUG)
    else:  # Default to WARNING, following Python logging standards
        logger.setLevel(logging.WARNING)


def _add_log_handler_with_formatter(handler: logging.Handler, name:str = None):
    """Add a handler to a logger.

    Args:
        handler (logging.Handler): A logging handler.
        name (str, optional): Which logger to configure. Default to root logger.
    """
    handler.setFormatter(_logging_format)
    logger = logging.getLogger(name)
    logger.addHandler(handler)



"""Functions to help tracing the application, i.e. via logging.
"""
import logging
import threading
from typing import Callable

# If humanize package is installed, use it to format bytes
try:
    from humanize import naturalsize
except ImportError:  # TODO: It is bad-practice to catch ImportErrors... Fix TBD
    def naturalsize(num: int) -> str:
        return f'{num} Bytes'

class DownloadTracer:
    """Log meta-info about any long-running download task.

    Note: This is Thread-safe -- can be used with stdlib "`threading`" module.
        Downloads are I/O bound anyhow, so multithreading works well.
    """

    def __init__(
        self, service_name: str, threshold_bytes: int = 10000, log_function: Callable = logging.info
    ):
        self._name = service_name
        self._threshold = threshold_bytes
        self._total_bytes_downloaded = 0
        self._prior_total_bytes_downloaded = 0
        self._lock = threading.Lock()
        self._lock_timeout = 0.1
        self._log_function = log_function


    @property
    def total_bytes(self):
        return self._total_bytes_downloaded

    def trace(self, data_len: int):
        """Log message for every  data has been Add data_len to the running total."""
        if self._lock.acquire(blocking=True, timeout=1):
            try:
                self._total_bytes_downloaded += data_len
                if (
                    self._total_bytes_downloaded >   self._threshold 
                                                + self._prior_total_bytes_downloaded
                ):
                    self._prior_total_bytes_downloaded = self._total_bytes_downloaded
                    self._log_downloaded_bytes()
            finally:
                self._lock.release()
        else:
            pass
        self._asyncio_loop

    def _log_downloaded_bytes(self):
        """Log the total bytes downloaded from this service."""
        bytes_downloaded_str = f'{naturalsize(self._total_bytes_downloaded)} Bytes'
        self._log_function(
            f'Total downloaded from {self._name}: {bytes_downloaded_str}'
        )



import logging
import time

import uologging
from assertpy import assert_that


def test_trace(caplog):
    # Arrange
    execution_time = 0.1
    with caplog.at_level(logging.DEBUG):
        uologging.init_console()
        logger = logging.getLogger(__name__)

        @uologging.trace(logger)
        def test_method():
            time.sleep(execution_time)

        # Act
        test_method()

    # Assert - May fail on very slow computers...
    assert_that(caplog.text).contains(f'Starting: {__name__}:{test_method.__name__}')
    assert_that(caplog.text).contains(f'Finished: {__name__}:{test_method.__name__}')
    assert_that(caplog.text).contains(f'exec time: {execution_time}')


def test_trace_dont_capture_secret_arguments(caplog):
    # Arrange
    with caplog.at_level(logging.DEBUG):
        uologging.init_console()
        logger = logging.getLogger(__name__)

        @uologging.trace(logger, capture_args=False)
        def test_method(password):
            pass

        # Act
        test_method('SuperSecretPassword')

    # Assert
    assert_that(caplog.text).does_not_contain('SuperSecretPassword')


def test_trace_capture_arguments(caplog):
    # Arrange
    with caplog.at_level(logging.DEBUG):
        uologging.init_console()
        logger = logging.getLogger(__name__)

        @uologging.trace(logger)
        def test_method(arg, another_arg=None):
            pass

        # Act
        test_method('Bird', 
                    another_arg=['is', 'the', 'word!'])

    # Assert
    assert_that(caplog.text).contains("test_method('Bird', another_arg=['is', 'the', 'word!'])")



import importlib

import example.hello
import example2.hello
import pytest
import uologging
from assertpy import assert_that


@pytest.fixture(autouse=True)
def reload_uologging_module_before_each_test_run():
    # uologging module has some module-variables that we need reset between each test run
    importlib.reload(uologging.uologging)
    yield


def test_init_console(caplog):
    # Arrange
    uologging.init_console('example')

    # Act -- do things that cause logging to occur
    example.hello.hello()

    # Assert
    assert_that(caplog.text).contains('WARNING')
    assert_that(caplog.text).does_not_contain('INFO')
    assert_that(caplog.text).does_not_contain('DEBUG')


@pytest.mark.parametrize('verbosity,expected_log_msg', 
                         [
                             (0, 'WARNING'),
                             (1, 'INFO'),
                             (2, 'DEBUG'),
                         ]
                         )
def test_set_verbosity(verbosity, expected_log_msg, caplog):
    # Arrange
    uologging.init_console('example')
    uologging.set_verbosity(verbosity, 'example')

    # Act -- do things that cause logging to occur
    example.hello.hello()

    # Assert
    assert_that(caplog.text).contains(expected_log_msg)


def test_init_console_2_packages(capsys):
    # Arrange
    uologging.init_console('example')
    uologging.set_verbosity(2, 'example')
    uologging.init_console('example2')
    uologging.set_verbosity(2, 'example2')

    # Act -- do things that cause logging to occur
    example.hello.hello()
    example2.hello.hello()

    # Assert
    logs = capsys.readouterr().err
    assert_that(logs).contains('Starting: example.hello:hello')
    assert_that(logs).contains('Starting: example2.hello:hello')


def test_init_console_package_twice(capsys):
    # Arrange
    uologging.init_console('example')
    uologging.set_verbosity(2, 'example')
    uologging.init_console('example')
    uologging.set_verbosity(2, 'example')

    # Act -- do things that cause logging to occur
    example.hello.hello()

    # Assert
    logs = capsys.readouterr()
    logs = logs.err
    assert_that(logs).does_not_match(r'(?m)Starting.*\n*.*Starting')



from concurrent.futures import ThreadPoolExecutor

import pytest
from assertpy import assert_that
from uologging.downloads import DownloadTracer

log_message_count = 0
@pytest.mark.parametrize('request_count,request_bytes,threshold_bytes,expected_msg_count', [
    (1, 1, 0, 1),  # threshold_bytes=0 means log every request
    (1, 100, 1001, 0),
    (10, 100, 999, 1),
    (100, 100, 999, 10),
])
def test_download_tracer(request_count, request_bytes, threshold_bytes, expected_msg_count, caplog):
    # Arrange
    # A 'spy' function, to see what gets logged
    global log_message_count
    log_message_count = 0
    def spy_log_function(msg):
        global log_message_count
        log_message_count += 1
    download_tracer = DownloadTracer('MyService', threshold_bytes=threshold_bytes, log_function=spy_log_function)
    # Our custom 'http_get' function
    def http_get(url):
        download_tracer.trace(request_bytes)
    # A list of fake URLs to 'download'
    urls = [f'http://fake-url/{num}' for num in range(0, request_count)]

    # Act
    # Use ThreadPoolExecutor to concurrently download the URLs
    with ThreadPoolExecutor(max_workers=10) as executor:
        responses = executor.map(http_get, urls)

    # Assert
    assert_that(download_tracer.total_bytes).is_equal_to(request_count * request_bytes)
    assert_that(log_message_count).is_equal_to(expected_msg_count)

    # Cleanup
    log_message_count = 0
